#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
TomoPy example script to reconstruct TXM data set.

Refs:
https://dxchange.readthedocs.io/en/latest/source/demo/doc.demo.aps_32id.html
https://github.com/tomopy/tomopy
https://tomopy.readthedocs.io/en/latest/ipynb/tomopy.html
https://hub.mybinder.org/user/tomopy-tomopy-82a5yrww/notebooks/doc/demo/gridrec.ipynb
"""

from __future__ import print_function
import sys
import tomopy
import dxchange
import h5py

if __name__ == '__main__':

    # Get arguments.
    arguments = "<path to .hdf file> <start slice> <end slice> [<reconstruction directory>]"
    if len(sys.argv) < 4 or len(sys.argv) > 5:
       print("Arguments:", arguments)
       quit()
    fname = sys.argv[1]
    start_slice = int(sys.argv[2])
    end_slice = int(sys.argv[3])
    if len(sys.argv) == 5:
       rec_arg = sys.argv[4] + "/slice"
    else:
       rec_arg = "./slice"
       
    # Validate start slice.
    f = h5py.File(fname, 'r')
    max_slice = len(f['exchange']['data'][0]) - 1
    if start_slice > max_slice:
       print("Max start slice =", max_slice)
       quit()
    if end_slice > max_slice:
       print("Truncating end slice to maximum value:", max_slice)
       end_slice = max_slice

    # Reconstruct in batches to avoid memory exhaustion.
    batch_size = 100
    start_slice_work = start_slice
    end_slice_work = start_slice + batch_size - 1
    if end_slice_work > end_slice:
       end_slice_work = end_slice
    while start_slice_work <= end_slice:
        print("Reconstructing slice batch", start_slice_work, "to", end_slice_work)

        # Read APS 32-ID raw data.
        proj, flat, dark, theta = dxchange.read_aps_32id(fname, sino=(start_slice_work, end_slice_work + 1))
    
        # If data collection angles is not defined in the hdf file then set it as equally spaced between 0-180 degrees.
        if (theta is None):
            theta = tomopy.angles(proj.shape[0])
        else:
            pass
    
        # Flat-field correction of raw data.
        proj = tomopy.normalize(proj, flat, dark)
    
        # Find rotation center.
        rot_center = tomopy.find_center(proj, theta, ind=0, init=1024, tol=0.5)
        #print("Center of rotation: ", rot_center)
    
        # Reconstruct object using Gridrec algorithm.
        proj = tomopy.minus_log(proj)
        rec = tomopy.recon(proj, theta, center=rot_center, algorithm='gridrec')
    
        # Mask each reconstructed slice with a circle.
        rec = tomopy.circ_mask(rec, axis=0, ratio=0.95)
    
        # Write data as stack of TIFs.
        dxchange.write_tiff_stack(rec, rec_arg, None, 0, 5, start_slice_work, True)
        
        # Next batch.
        start_slice_work = end_slice_work + 1
        end_slice_work = end_slice_work + batch_size
        if end_slice_work > end_slice:
           end_slice_work = end_slice
